/*---------------------------------------------------------------------------*\

    Copyright (C) 2000 AVM GmbH. All rights reserved.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY, without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE, and WITHOUT
    ANY LIABILITY FOR ANY DAMAGES arising out of or in connection
    with the use or performance of this software. See the
    GNU General Public License for further details.

\*---------------------------------------------------------------------------*/

#include "ByteQueu.h"
#include "osImport.h"
#include <string.h>

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

#define ByteQueue_PutLock           0
#define ByteQueue_GetLock           1


/*===========================================================================*\
    CStandardByteQueue - Konstruktor
\*===========================================================================*/

CStandardByteQueue::CStandardByteQueue (tUInt maxQueueSize, tBool PartFillAllow) {
    dhead ("CStandardByteQueue-Constructor", DEBUG_CStandardByteQueue);
    dassert (maxQueueSize != 0);

    buffer   = new tUChar [maxQueueSize + 1];
    maxCount = (buffer) ? maxQueueSize + 1 : 0;
    head     = 0;
    tail     = 0;

    SetPartFill (PartFillAllow);
}


/*===========================================================================*\
    CStandardByteQueue - Destruktor
\*===========================================================================*/

CStandardByteQueue::~CStandardByteQueue (void) {
    dhead ("CStandardByteQueue-Destructor", DEBUG_CStandardByteQueue);
    tUChar *tmpBuf = buffer;
    ProtectPut.Begin();
    ProtectGet.Begin();
    maxCount = 0;
    head = 0;
    tail = 0;
    buffer = 0;
    ProtectGet.End();
    ProtectPut.End();
    delete [] tmpBuf;
}


/*===========================================================================*\
    CStandardByteQueue::Put
\*===========================================================================*/

tUInt CStandardByteQueue::Put (tUChar *dataIn, tUInt length) {
    dhead ("CStandardByteQueue::Put", DEBUG_CStandardByteQueue);
    dparams ("data=%p, len=%x", dataIn, length);
    dassert (dataIn != 0);
    dwarning (length != 0);
    dwarning (maxCount > 1);
    dwarning (buffer != 0);
    if (!maxCount) {
        return 0;
    }

    tUInt tail_free = tail;

    dprint ("h=%x,t=%x", head, tail);
    ProtectPut.Begin();
    if (tail_free <= head) {
        tail_free += maxCount;
    }
    tail_free -= head + 1;
    if (length > tail_free) {
        dprint ("Queue full: in=%x, free=%x!", length, tail_free);
        length = (bPartFill == vTrue) ? tail_free : 0;
    }
    if (length) {
        if (length + head >= maxCount) {
            tUInt dif = maxCount - head;
            memcpy (buffer + head, dataIn, dif);
            head = length - dif;
            memcpy (buffer, dataIn + dif, head);
        } else {
            memcpy (buffer + head, dataIn, length);
            head += length;
        }
    }
    ProtectPut.End();
    dprint ("h=%x,t=%x", head, tail);
    RETURN ('x', length);
}


/*===========================================================================*\
    CStandardByteQueue::PutInFront
\*===========================================================================*/

tUInt CStandardByteQueue::PutInFront (tUChar *dataIn, tUInt length) {
    dhead ("CStandardByteQueue::PutInFront", DEBUG_CStandardByteQueue);
    dparams ("data=%p, len=%x", dataIn, length);
    dassert (dataIn != 0);
    dwarning (length != 0);
    dwarning (maxCount > 1);
    dwarning (buffer != 0);
    if (!maxCount) {
        return 0;
    }
    tUInt free;
    tUInt offset = 0;

    dprint ("h=%x,t=%x", head, tail);
    ProtectPut.Begin();
    ProtectGet.Begin();
    free = ((tail <= head) ? maxCount : 0) + tail - head - 1;
    if (length > free) {
        dprint ("Queue full: in=%x, free=%x!", length, free);
        offset = length - free;
        length = (bPartFill == vTrue) ? free : 0;
    }

    if (tail < length) {
        tUInt dif = length - tail;
        memcpy (buffer, dataIn + offset + dif, tail);
        tail = maxCount - dif;
        memcpy (buffer + tail, dataIn + offset, dif);
    } else {
        memcpy (buffer + tail - length, dataIn + offset, length);
        tail -= length;
    }
    ProtectGet.End();
    ProtectPut.End();
    dprint ("h=%x,t=%x", head, tail);
    RETURN ('x', length);
}


/*===========================================================================*\
    CStandardByteQueue::Get
\*===========================================================================*/

tUInt CStandardByteQueue::Get (tUChar *dataOut, tUInt maxLength) {
    dhead ("CStandardByteQueue::Get", DEBUG_CStandardByteQueue);
    dparams ("data=%p, len=%x", dataOut, maxLength);
    dassert (dataOut != 0);
    dwarning (maxLength > 0);
    dwarning (maxCount > 1);
    dwarning (buffer != 0);
    tUInt head_count = head;        

    dprint ("h=%x,t=%x", head, tail);
    ProtectGet.Begin();
    if (head_count < tail) {
        head_count += maxCount;
    }
    head_count -= tail;
    if (maxLength > head_count) {
        dprint ("Not enough data: out=%x, use=%x!", maxLength, head_count);
        maxLength = (bPartFill == vTrue) ? head_count : 0;
    }
    if (maxLength + tail >= maxCount) {
        tUInt dif = maxCount - tail;
        memcpy (dataOut, buffer + tail, dif);
        tail = maxLength - dif;
        memcpy (dataOut + dif, buffer, tail);
    } else {
        memcpy (dataOut, buffer + tail, maxLength);
        tail += maxLength;
    }
    ProtectGet.End();
    dprint ("h=%x,t=%x", head, tail);
    RETURN ('x', maxLength);
}


/*===========================================================================*\
    CStandardByteQueue::Clear
\*===========================================================================*/

void CStandardByteQueue::Clear (void) {
    dhead ("CStandardByteQueue::Clear", DEBUG_CStandardByteQueue);
    dwarning (buffer != 0);

    ProtectPut.Begin();
    ProtectGet.Begin();
    head = 0;
    tail = 0;
    ProtectGet.End();
    ProtectPut.End();
}


/*===========================================================================*\
    CStandardByteQueue::Resize
\*===========================================================================*/

tBool CStandardByteQueue::Resize (tUInt newMaxQueueSize) {
    dhead ("CStandardByteQueue::Resize", DEBUG_CStandardByteQueue);
    dparams ("newSize=%x", newMaxQueueSize);
    dassert (newMaxQueueSize > 0);
    tUChar *newBuf = new tUChar [newMaxQueueSize + 1];
    tUChar *tmpBuf = newBuf;
    tBool   fret   = vFalse;

    if (newBuf) {
        ProtectPut.Begin();
        ProtectGet.Begin();
        tUInt count = ((head < tail) ?  maxCount : 0) + head - tail;
        if (count <= newMaxQueueSize) {
            if (head < tail) {
                tUInt dif = maxCount - tail;
                memcpy (newBuf, buffer + tail, dif);
                memcpy (newBuf + dif, buffer, head);
            } else {
                memcpy (newBuf, buffer + tail, head - tail);
            }
            tmpBuf = buffer;
            buffer = newBuf;
            head = count;
            tail = 0;
            maxCount = newMaxQueueSize + 1;
            fret = vTrue;
        }
        ProtectGet.End();
        ProtectPut.End();
        delete [] tmpBuf;
    }
    RETURN ('x', fret);
}


/*===========================================================================*\
\*===========================================================================*/
