/*---------------------------------------------------------------------------*\

    Copyright (C) 2000 AVM GmbH. All rights reserved.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY, without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE, and WITHOUT
    ANY LIABILITY FOR ANY DAMAGES arising out of or in connection
    with the use or performance of this software. See the
    GNU General Public License for further details.

\*---------------------------------------------------------------------------*/

#include <stdarg.h>
#include "DbgInfo.h"
#include "CDbgIDS.h"
#include "DbgStack.h"
#include "aStdLib.h"
#include "ExtFuncs.h"

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

#define DEBUGINFO_PRINTBUF_SIZE     100

/*===========================================================================*\
\*===========================================================================*/

#if (USE_DEBUG_INFO_CLASS > _USE_DEBUG_INFO_CLASS_NONE)

#define dwrite_MAX_CHAR_LENGTH      150

void dwriteE (unsigned DL, char *format, ...) {
    if (DL >= Dbg_Level_Short) {
        char string[dwrite_MAX_CHAR_LENGTH];
        va_list param;
        va_start (param, format);
        if (a_vsnprintf (string, dwrite_MAX_CHAR_LENGTH, format, param) >= 0) {
            DbgIPrintStr (string);
        }
        va_end (param);
    }
}

void dwriteI (unsigned DL, char *format, ...) {
    if (DL >= Dbg_Level_Internal) {
        char string[dwrite_MAX_CHAR_LENGTH];
        va_list param;
        va_start (param, format);
        if (a_vsnprintf (string, dwrite_MAX_CHAR_LENGTH, format, param) >= 0) {
            DbgIPrintStr (string);
        }
        va_end (param);
    }
}

#endif

/*===========================================================================*\
\*===========================================================================*/
#if (USE_DEBUG_INFO_CLASS == _USE_DEBUG_INFO_CLASS_EMULATION)

void dparams (const char *format, ...) {
    va_list param;
    va_start (param, format);
    char output[DBGDYNSTRING_STARTMAXLENGTH];
    if (a_vsnprintf(output, DBGDYNSTRING_STARTMAXLENGTH, format, param) >= 0) {
        DbgIPrintFull (PARAMETERS_BEGIN "%s" PARAMETERS_END, output);
    } else {
        DbgIPrintFull ("\ndparams: buffer to small\n");
    }
    va_end (param);
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

void dprintBuf (const void *data, unsigned length) {
    if (data && length) {
        unsigned       i;
        unsigned       j;
        unsigned       laufbuffree;
        char           outstring[DEBUGINFO_PRINTBUF_SIZE];
        char          *laufbuf;
        unsigned char *databuf;

        for (i = 0; i < length; i += 0x10) {
            a_snprintf (outstring, DEBUGINFO_PRINTBUF_SIZE, "\n%05x: ", (i & 0xFFFFF));
            laufbuffree = DEBUGINFO_PRINTBUF_SIZE - 8;
            laufbuf = outstring + 8;
            databuf = (((unsigned char *)data) + i);
            for (j = 0; j < 0x10; j++) {
                if (i + j < length) {
                    if (j && (j % 4 == 0)) {
                        DbgIAssert (laufbuffree >= 5);
                        a_snprintf (laufbuf, laufbuffree, "' %02x ", *databuf);
                        laufbuffree -= 5;
                        laufbuf     += 5;
                    } else {
                        DbgIAssert (laufbuffree >= 3);
                        a_snprintf (laufbuf, laufbuffree, "%02x ", *databuf);
                        laufbuf += 3;
                    }
                    databuf++;
                } else {
                    if (j % 4 == 0) {
                        DbgIAssert (laufbuffree >= 5);
                        a_snprintf (laufbuf, laufbuffree, "     ");
                        laufbuf += 5;
                    } else {
                        DbgIAssert (laufbuffree >= 3);
                        a_snprintf (laufbuf, laufbuffree, "   ");
                        laufbuf += 3;
                    }
                }
            }
            DbgIAssert (laufbuffree >= 20);
            *(laufbuf++) = '|';
            *(laufbuf++) = ' ';
            databuf = (((unsigned char *)data) + i);
            for (j = 0; ((j < 0x10) && (i + j < length)); j++, laufbuf++, databuf++) {
                *laufbuf = (*databuf < 32) ? '.' : *databuf;
            }
            *(laufbuf++) = ' ';
            *laufbuf     = '\0';
            DbgIPrintFull (outstring);
        }
    } else {
        DbgIPrintFull ("DATA: (p=%x, l=%x)\n", data, length);
    }
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

int dinfo (const int expression, const char *format, ...) {
    int fret = 0;
    if (!expression) {
        va_list param;
        va_start (param, format);
        char output[DBGDYNSTRING_STARTMAXLENGTH];
        fret = a_vsnprintf(output, DBGDYNSTRING_STARTMAXLENGTH, format, param);
        if (fret >= 0) {
            DbgIPrintFull ("[info: %s]\n", output);
        }
        va_end (param);
    }
    return fret;
}


/*===========================================================================*\
\*===========================================================================*/
#elif (USE_DEBUG_INFO_CLASS >= _USE_DEBUG_INFO_CLASS_MINIMAL)


#ifdef DEBUG_INDENT_WIDTH
static unsigned IndentSpaces = 0;
#endif

#if (USE_DEBUG_INFO_CLASS >= _USE_DEBUG_INFO_CLASS_FULL)
static CDebugStack DebugInfoCallStack;
#endif


/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

CDebugInfo::CDebugInfo(char *FuncName, unsigned DebugLevel, unsigned PrintLevel,
                       char *_FileName, unsigned _LineNumber) {
    RetValue.format = 0;
    debuglevel = DebugLevel;
    if (DebugLevel > Dbg_Level_Nothing) {
        #ifdef DEBUG_INDENT_WIDTH
            spaces = IndentSpaces;
            IndentSpaces += DEBUG_INDENT_WIDTH;
        #else
            spaces = 0;
        #endif

        name = FuncName;
        linenumber = _LineNumber;
        printlevel = PrintLevel;

        filename = _FileName;
        for (; *_FileName != '\0'; _FileName++) {
            if ((*_FileName == '\\') || (*_FileName == '/')) {
                filename = _FileName + 1;
            }
        }

        #if (USE_DEBUG_INFO_CLASS >= _USE_DEBUG_INFO_CLASS_FULL)
            hCallStack = DebugInfoCallStack.AddTo (this);
        #endif

        if (printlevel & Prt_Level_NameAtFuncStart) {
            if (debuglevel == Dbg_Level_Short) {
                RememberOutput (NAMEFORSHORT_BEGIN "%s", name);
                DbgPrintIntern (Prt_Level_Mask_NoMore);
            } else {
                #if (USE_DEBUG_INFO_CLASS == _USE_DEBUG_INFO_CLASS_MINIMAL)
                    if (debuglevel >= Dbg_Level_Calls) {
                #endif
                if (printlevel & Prt_Level_FileNameAtStart) {
                    #ifdef DEBUG_INDENT_WIDTH
                        RememberOutput ("%*t" NAMEATFUNCSTART_BEGIN "%s in %s(%d)" NAMEATFUNCSTART_END,
                                        spaces, ' ', name, filename, linenumber);
                    #else
                        RememberOutput (NAMEATFUNCSTART_BEGIN "%s in %s(%d)" NAMEATFUNCSTART_END,
                                        name, filename, linenumber);
                    #endif
                } else {
                    #ifdef DEBUG_INDENT_WIDTH
                        RememberOutput ("%*t" NAMEATFUNCSTART_BEGIN "%s" NAMEATFUNCSTART_END,
                                        spaces, ' ', name);
                    #else
                        RememberOutput (NAMEATFUNCSTART_BEGIN "%s" NAMEATFUNCSTART_END, name);
                    #endif
                }
                #if (USE_DEBUG_INFO_CLASS == _USE_DEBUG_INFO_CLASS_MINIMAL)
                    }
                #else
                    if (debuglevel >= Dbg_Level_Calls) {
                        DbgPrintIntern (Prt_Level_Mask_NoMore);
                    }
                #endif
            }
        }
    } else {
        name       = 0;
        filename   = 0;
        linenumber = 0;
        printlevel = 0;
        spaces     = 0;
        #if (USE_DEBUG_INFO_CLASS >= _USE_DEBUG_INFO_CLASS_FULL)
        hCallStack = 0;
        #endif
    }
}


/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

CDebugInfo::~CDebugInfo (void) {
    /*----- Ende der Funktion ausgeben -----*/
    if (debuglevel == Dbg_Level_Short) {
        RememberOutput (NAMEFORSHORT_END, name);
        DbgPrintIntern (Prt_Level_Mask_NoMore);
    } else if ((debuglevel >= Dbg_Level_Calls) && (printlevel & Prt_Level_NameAtFuncStart)) {
        if ((debuglevel >= Dbg_Level_Entry) && (RetValue.format != '\0')) {
            char ParamOutput[sizeof(NAMEATFUNCEND_BEGIN) + sizeof(NAMEATFUNCEND_END) +
                             sizeof(RETURNVALUE_BEGIN) + sizeof(RETURNVALUE_END) + 8]
            #ifdef DEBUG_INDENT_WIDTH
                = "%*t" NAMEATFUNCEND_BEGIN RETURNVALUE_BEGIN "%." RETURNVALUE_END "%s%*t" NAMEATFUNCEND_END;

                ParamOutput[sizeof(NAMEATFUNCEND_BEGIN) +sizeof(RETURNVALUE_BEGIN) +2] = RetValue.format;
                RememberOutput (ParamOutput, spaces, ' ', RetValue.p, name,
                                spaces - DEBUG_INDENT_WIDTH, ' ');
            #else
                = NAMEATFUNCEND_BEGIN RETURNVALUE_BEGIN "%." RETURNVALUE_END "%s" NAMEATFUNCEND_END;

                ParamOutput[sizeof (NAMEATFUNCEND_BEGIN) +sizeof (RETURNVALUE_BEGIN) -1] = RetValue.format;
                RememberOutput (ParamOutput, RetValue.p, name);
            #endif
        } else {
            #ifdef DEBUG_INDENT_WIDTH
                RememberOutput ("%*t" NAMEATFUNCEND_BEGIN "%s%*t" NAMEATFUNCEND_END,
                                spaces, ' ', name, spaces - DEBUG_INDENT_WIDTH, ' ');
            #else
                RememberOutput (NAMEATFUNCEND_BEGIN "%s" NAMEATFUNCEND_END, name);
            #endif
        }
        DbgPrintIntern (Prt_Level_Mask_NoMore);
    }

    #if (USE_DEBUG_INFO_CLASS >= _USE_DEBUG_INFO_CLASS_FULL)
        DebugInfoCallStack.Remove (hCallStack);
    #endif

    #ifdef DEBUG_INDENT_WIDTH
    if (debuglevel > Dbg_Level_Nothing) {
        IndentSpaces -= DEBUG_INDENT_WIDTH;
    }
    #endif

    /*----- Speicherplatz aufrumen -----*/
    name = 0;
    filename = 0;
    linenumber = 0;
    debuglevel = printlevel = 0;
    spaces = 0;
    #if (USE_DEBUG_INFO_CLASS >= _USE_DEBUG_INFO_CLASS_FULL)
        hCallStack = 0;
    #endif
    RetValue.format = 0;
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

int CDebugInfo::print (const char *format, ...) {
    if (name == 0) return 0;

    va_list param;
    va_start (param, format);
    int fret;
    #if (USE_DEBUG_INFO_CLASS == _USE_DEBUG_INFO_CLASS_MINIMAL)
        fret = (debuglevel >= Dbg_Level_Internal) ? RememberOutput ("%a", format, param) : 0;
    #else
        fret = RememberOutput ("%a", format, param);
        if (debuglevel >= Dbg_Level_Internal) {
            DbgPrintIntern (Prt_Level_Mask_NoMore);
        }
    #endif
    va_end (param);
    return fret;
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

void CDebugInfo::printBuf (const void *data, unsigned length) {
    if (data && length && (debuglevel >= Dbg_Level_Internal)) {
        unsigned       i;
        unsigned       j;
        unsigned       laufbuffree;
        char           outstring[DEBUGINFO_PRINTBUF_SIZE];
        char          *laufbuf;
        unsigned char *databuf;

        for (i = 0; i < length; i += 0x10) {
            a_snprintf (outstring, DEBUGINFO_PRINTBUF_SIZE, "\n%05x: ", (i & 0xFFFFF));
            laufbuffree = DEBUGINFO_PRINTBUF_SIZE - 8;
            laufbuf = outstring + 8;
            databuf = (((unsigned char *)data) + i);
            for (j = 0; j < 0x10; j++) {
                if (i + j < length) {
                    if (j && (j % 4 == 0)) {
                        DbgIAssert (laufbuffree >= 5);
                        a_snprintf (laufbuf, laufbuffree, "' %02x ", *databuf);
                        laufbuffree -= 5;
                        laufbuf     += 5;
                    } else {
                        DbgIAssert (laufbuffree >= 3);
                        a_snprintf (laufbuf, laufbuffree, "%02x ", *databuf);
                        laufbuf += 3;
                    }
                    databuf++;
                } else {
                    if (j % 4 == 0) {
                        DbgIAssert (laufbuffree >= 5);
                        a_snprintf (laufbuf, laufbuffree, "     ");
                        laufbuf += 5;
                    } else {
                        DbgIAssert (laufbuffree >= 3);
                        a_snprintf (laufbuf, laufbuffree, "   ");
                        laufbuf += 3;
                    }
                }
            }
            DbgIAssert (laufbuffree >= 20);
            *(laufbuf++) = '|';
            *(laufbuf++) = ' ';
            databuf = (((unsigned char *)data) + i);
            for (j = 0; ((j < 0x10) && (i + j < length)); j++, laufbuf++, databuf++) {
                *laufbuf = (*databuf < 32) ? '.' : *databuf;
            }
            *(laufbuf++) = ' ';
            *laufbuf     = '\0';
            RememberOutput (outstring);
            DbgPrintIntern (Prt_Level_Mask_NoMore);
        }
    } else {
        #if (USE_DEBUG_INFO_CLASS == _USE_DEBUG_INFO_CLASS_MINIMAL)
            if (debuglevel >= Dbg_Level_Internal) {
                RememberOutput ("DATA: (p=%p, l=%x)\n", data, length);
            }
        #else
            RememberOutput ("DATA: (p=%x, l=%x)\n", data, length);
            if (debuglevel >= Dbg_Level_Internal) {
                DbgPrintIntern (Prt_Level_Mask_NoMore);
            }
        #endif
    }
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

int CDebugInfo::printWrn (const char *format, ...) {
    if (name == 0) return 0;

    va_list param;
    va_start (param, format);
    int fret = 0;
    #if (USE_DEBUG_INFO_CLASS == _USE_DEBUG_INFO_CLASS_MINIMAL)
        if (debuglevel >= Dbg_Level_Infos) {
    #endif
    fret = (printlevel & Prt_Level_FileNameAtInfo) ?
            RememberOutput ("[NOTE(%s in %s(%d+)): %a]\n", name, filename, linenumber, format, param) :
            RememberOutput ("[NOTE(%s): %a]\n", name, format, param);
    #if (USE_DEBUG_INFO_CLASS == _USE_DEBUG_INFO_CLASS_MINIMAL)
        }
    #else
        if (debuglevel >= Dbg_Level_Infos) {
            DbgPrintIntern (printlevel & 0x03);
        }
    #endif
    va_end (param);
    return fret;
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

int CDebugInfo::printErr (const char *format, ...) {
    if (name == 0) return 0;

    va_list param;
    va_start (param, format);
    int fret = 0;
    #if (USE_DEBUG_INFO_CLASS == _USE_DEBUG_INFO_CLASS_MINIMAL)
        if (debuglevel >= Dbg_Level_Errors) {
    #endif
    fret = (printlevel & Prt_Level_FileNameAtInfo) ?
            RememberOutput ("[ERROR(%s in %s(%d+)): %a]\n", name, filename, linenumber, format, param) :
            RememberOutput ("[ERROR(%s): %a]\n", name, format, param);
    #if (USE_DEBUG_INFO_CLASS == _USE_DEBUG_INFO_CLASS_MINIMAL)
        }
    #else
        if (debuglevel >= Dbg_Level_Errors) {
            DbgPrintIntern (printlevel & 0x03);
        }
    #endif
    va_end (param);
    return fret;
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

void CDebugInfo::assertion (const char *exp, const unsigned int line) {
    if (!name) return;

    #if (USE_DEBUG_INFO_CLASS == _USE_DEBUG_INFO_CLASS_MINIMAL)
        if (debuglevel >= Dbg_Level_Errors) {
    #endif
    if (printlevel & Prt_Level_FuncNameAtError) {
        RememberOutput ("[assertion failed: (%s) %s in %s(%d)]\n", exp, name, filename, line);
    } else {
        RememberOutput ("[assertion failed: (%s) in %s(%d)]\n", exp, filename, line);
    }
    #if (USE_DEBUG_INFO_CLASS == _USE_DEBUG_INFO_CLASS_MINIMAL)
        }
    #else
        if (debuglevel >= Dbg_Level_Errors) {
            DbgPrintIntern (printlevel & 0x03);
        }
    #endif
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

void CDebugInfo::assertion (const int expValue, const char *exp, const char *fname,
                            const unsigned int line) {
    if (expValue || !name) return;
    unsigned last = 0;
    for (unsigned i = 0; fname[i] != '\0'; i++) {
        if ((fname[i] == '\\') || (fname[i] =='/')) {
            last = i + 1;
        }
    }
    fname += last;
    #if (USE_DEBUG_INFO_CLASS == _USE_DEBUG_INFO_CLASS_MINIMAL)
        if (debuglevel >= Dbg_Level_Errors) {
    #endif
    if (printlevel & Prt_Level_FuncNameAtError) {
        RememberOutput ("[assertion failed: (%s) %s in %s(%d)]\n", exp, name, fname, line);
    } else {
        RememberOutput ("[assertion failed: (%s) in %s(%d)]\n", exp, fname, line);
    }
    #if (USE_DEBUG_INFO_CLASS == _USE_DEBUG_INFO_CLASS_MINIMAL)
        }
    #else
        if (debuglevel >= Dbg_Level_Errors) {
            DbgPrintIntern (printlevel & 0x03);
        }
    #endif
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

void CDebugInfo::warning (const char *exp, const unsigned int line) {
    if (!name) return;

    #if (USE_DEBUG_INFO_CLASS == _USE_DEBUG_INFO_CLASS_MINIMAL)
        if (debuglevel >= Dbg_Level_Infos) {
    #endif
    if (printlevel & Prt_Level_FuncNameAtError) {
        RememberOutput ("[warning: (%s) %s in %s(%d)]\n", exp, name, filename, line);
    } else {
        RememberOutput ("[warning: (%s) in %s(%d)]\n", exp, filename, line);
    }
    #if (USE_DEBUG_INFO_CLASS == _USE_DEBUG_INFO_CLASS_MINIMAL)
        }
    #else
        if (debuglevel >= Dbg_Level_Infos) {
            DbgPrintIntern ((printlevel & 0x0c) >> 2);
        }
    #endif
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

void CDebugInfo::warning (const int expValue, const char *exp, const char *fname,
                          const unsigned int line) {
    if (expValue || !name) return;
    unsigned last = 0;
    for (unsigned i = 0; fname[i] != '\0'; i++) {
        if ((fname[i] == '\\') || (fname[i] =='/')) {
            last = i + 1;
        }
    }
    fname += last;
    #if (USE_DEBUG_INFO_CLASS == _USE_DEBUG_INFO_CLASS_MINIMAL)
        if (debuglevel >= Dbg_Level_Infos) {
    #endif
    if (printlevel & Prt_Level_FuncNameAtError) {
        RememberOutput ("[warning: (%s) %s in %s(%d)]\n", exp, name, fname, line);
    } else {
        RememberOutput ("[warning: (%s) in %s(%d)]\n", exp, fname, line);
    }
    #if (USE_DEBUG_INFO_CLASS == _USE_DEBUG_INFO_CLASS_MINIMAL)
        }
    #else
        if (debuglevel >= Dbg_Level_Infos) {
            DbgPrintIntern ((printlevel & 0x0c) >> 2);
        }
    #endif
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

int CDebugInfo::info (const int expression, const char *format, ...) {
    if (expression || !name) return 0;
    va_list param;
    va_start (param, format);
    int fret = 0;
    #if (USE_DEBUG_INFO_CLASS == _USE_DEBUG_INFO_CLASS_MINIMAL)
        if (debuglevel >= Dbg_Level_Infos) {
    #endif
    fret = (printlevel & Prt_Level_FileNameAtInfo) ?
            RememberOutput ("[info(%s in %s(%d+)): %a]\n", name, filename, linenumber, format, param) :
            RememberOutput ("[info(%s): %a]\n", name, format, param);
    #if (USE_DEBUG_INFO_CLASS == _USE_DEBUG_INFO_CLASS_MINIMAL)
        }
    #else
        if (debuglevel >= Dbg_Level_Infos) {
            DbgPrintIntern ((printlevel & 0x30) >> 4);
        }
    #endif
    va_end (param);
    return fret;
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

void CDebugInfo::PrintParams (const char *format, ...) {
    if (name == 0) return;

    va_list param;
    va_start (param, format);
    if (debuglevel == Dbg_Level_Short) {
        RememberOutput (PARAMETERS_SHORT_BEGIN "%a" PARAMETERS_SHORT_END, format, param);
        DbgPrintIntern (Prt_Level_Mask_NoMore);
    } else {
        #if (USE_DEBUG_INFO_CLASS == _USE_DEBUG_INFO_CLASS_MINIMAL)
            if (debuglevel >= Dbg_Level_Entry) {
                RememberOutput (PARAMETERS_BEGIN "%a" PARAMETERS_END, format, param);
            }
        #else
            RememberOutput (PARAMETERS_BEGIN "%a" PARAMETERS_END, format, param);
            if (debuglevel >= Dbg_Level_Entry) {
                DbgPrintIntern (Prt_Level_Mask_NoMore);
            }
        #endif
    }
    va_end (param);
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

void CDebugInfo::SetRetVal (const char format, ...) {
    va_list param;
    va_start (param, format);

    RetValue.format = format;
    // Lieber einen kleine Hack, weil schneller !!!
    // Es wird sich der speicher-grte Parameter gemerkt und
    // erst bei der Ausgabe die Fallabfrage gemacht.
    RetValue.p = va_arg (param, void *);
    va_end (param);
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

void CDebugInfo::ShowAllOutputs (void) {
    #if (USE_DEBUG_INFO_CLASS > _USE_DEBUG_INFO_CLASS_MINIMAL)
        if (!Output.IsEmpty()) {
            char *outstr = Output.GetAll();
            DbgIPrintStr (outstr);
        }
        /*----- TODO: Funktionsnamen ausgeben, immer oder wennn nichts in der OutputQueue -----*/
    #endif
}

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

int CDebugInfo::RememberOutput (const char *format, ...) {
    int fret;

    va_list param;
    va_start (param, format);
    //DbgIPrintStr ("In RememberOutput ");
    #if (USE_DEBUG_INFO_CLASS > _USE_DEBUG_INFO_CLASS_MINIMAL)
        fret = (printlevel & Prt_Level_NameInEveryLine) ?
            #ifdef DEBUG_INDENT_WIDTH
                    Output.AddTo ("%*t" NAMEINEVERYLINE_BEGIN "%s" NAMEINEVERYLINE_END "%a",
                                  spaces, ' ', name, format, param) :
                    Output.AddTo ("%*t%a", spaces, ' ', format, param);
            #else
                    Output.AddTo (NAMEINEVERYLINE_BEGIN "%s" NAMEINEVERYLINE_END "%a", name, format, param):
                    Output.AddTo (format, param);
            #endif
    #else
        char output[DBGDYNSTRING_STARTMAXLENGTH];
        fret = (printlevel & Prt_Level_NameInEveryLine) ?
            #ifdef DEBUG_INDENT_WIDTH
                a_snprintf (output, DBGDYNSTRING_STARTMAXLENGTH,
                            "%*t" NAMEINEVERYLINE_BEGIN "%s" NAMEINEVERYLINE_END "%a",
                            spaces, ' ', name, format, param) :
                a_snprintf (output, DBGDYNSTRING_STARTMAXLENGTH, "%*t%a", spaces, ' ', format, param);
            #else
                a_snprintf (output, DBGDYNSTRING_STARTMAXLENGTH,
                            NAMEINEVERYLINE_BEGIN "%s" NAMEINEVERYLINE_END "%a",
                            name, format, param) :
                a_vsnprintf (output, DBGDYNSTRING_STARTMAXLENGTH, format, param);
            #endif
        if (fret >= 0) {
            DbgIPrintStr (output);
        }
    #endif
    va_end (param);
    return fret;
}


/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/
#if (USE_DEBUG_INFO_CLASS > _USE_DEBUG_INFO_CLASS_MINIMAL)

void CDebugInfo::DbgPrintIntern (unsigned PrtLevelMask) {
    char *outstr;

    switch (PrtLevelMask) {
    case Prt_Level_Mask_AllFunc:
        #if (USE_DEBUG_INFO_CLASS >= _USE_DEBUG_INFO_CLASS_FULL)
            outstr = Output.GetLastEntry();
            DbgIPrintStr (outstr);
            if ((DebugInfoCallStack.CountItems() > 1) || !Output.IsEmpty()) {
                DbgIPrintStr ("\n----- All Previous Outputs -------------------------------------------\n");
                DebugInfoCallStack.ShowAllDebugOutputs();
                DbgIPrintStr ("\n----------------------------------------------------------------------\n\n\n");
            }
            break;
        #endif
    case Prt_Level_Mask_CurFunc:
        outstr = Output.GetLastEntry();
        DbgIPrintStr (outstr);
        if (!Output.IsEmpty()) {
            DbgIPrintStr ("\n----- Previous Output ------------------------------------------------\n");
            outstr = Output.GetAll();
            DbgIPrintStr (outstr);
            DbgIPrintStr ("\n----------------------------------------------------------------------\n\n");
        }
        break;
    case Prt_Level_Mask_NoMore:
        outstr = Output.GetLastEntry();
        DbgIPrintStr (outstr);
        Output.Delete();
        break;
    default:
        /*----- Illegal PrtLevelMask!! -----*/
        break;
    }
}

#endif /* #if (USE_DEBUG_INFO_CLASS > _USE_DEBUG_INFO_CLASS_MINIMAL) */


/*===========================================================================*\
\*===========================================================================*/
#endif /* #elif (USE_DEBUG_INFO_CLASS >= _USE_DEBUG_INFO_CLASS_NORMAL) */



